#include "usecurses.h"
#include "connect.h"

#define DefaultPort 5819
#define ShipCount 13

static const char *HostName = NULL;
static int Port = DefaultPort;
static int Sock;

typedef char Field[10][10];

static struct Laiva
{
	unsigned char x, y;
	signed char LenDir;
	/* Length = Abs(LenDir), (LenDir<0)?Vertical:Horizontal */
} Laivani[ShipCount];

static Field Oma, Toinen;

#define Tyhja 0		/* Kartalla: Tyhj alue                            */
#define Laiva 1		/* Kartalla: laiva jota ei ole lydetty            */
#define End   0		/* Protokolla: Lopeta peli                         */
#define Hit   2		/* Protokolla: Osuit, kartalla: Osuttu             */
#define Uppo  3		/* Protokolla: Upotit, kartalla: Opotettu          */
#define Vuoro 4     /* Protokolla: Sun vuorosi                         */
#define Idle  0x20  /* Protokolla: Tm skipataaan; Kartalla: +Kursori */
#define Try   5		/* Protokolla: Rasti tnne; Kartalla: Huti         */

static unsigned char WantByte(void)
{
	unsigned char Byte;
	unsigned n = read(Sock, &Byte, 1);
	return Byte;
}

static void Kala(void)
{
	static const unsigned char Attrs[4] = {0x19,0x1B,0x13,0x17};
	static const char Kalat[8][3] = {"~ "," ~","  ","  ","  ","  ","  ","  "};
	
	SetAttr(Attrs[Random(4)]);
	printf(Kalat[Random(8)]);
}

static void UpdateField(Field *Kentta, int dx,int dy, int x1,int y1, int x2,int y2)
{
	int x, y;
	
	for(y=y1; y<=y2; y++)
	{
		printf("\33[%d;%dH", dy+y*2, dx + x1*2);
		
		if(y >= 0)
		{
			if((x=x1) == -1)
			{
				SetAttr(14);
				printf(" %c", 'A'+y);
				x++;
			}
		
			for(; x<=x2; x++)
				switch(Kentta[0][x][y]&~Idle)
				{
					case Tyhja: if(!(Kentta[0][x][y]&Idle)){Kala(); break;}
					case Uppo:  SetAttr(0x13); printf((Kentta[0][x][y]&Idle)?"\\/":"%%%%"); break;
					case Laiva: SetAttr(0x78); printf((Kentta[0][x][y]&Idle)?"\\/":"##"); break;
					case Hit:   SetAttr(0x47); printf((Kentta[0][x][y]&Idle)?"\\/":"@@"); break;
					case Try:   SetAttr(0x18); printf("\\/"); break;
				}
		}
		
		if((x=x1) == -1)x++;
		
		printf("\33[%d;%dH", dy+y*2+1, dx+x*2);
		
		for(; x<=x2; x++)
			if(y==-1)
			{
				SetAttr(14);
				printf("%-2d", x+1);
			}
			else
				switch(Kentta[0][x][y]&~Idle)
				{
					case Tyhja: if(!(Kentta[0][x][y]&Idle)){Kala(); break;}
					case Uppo:  SetAttr(0x13); printf((Kentta[0][x][y]&Idle)?"/\\":"%%%%"); break;
					case Laiva: SetAttr(0x78); printf((Kentta[0][x][y]&Idle)?"/\\":"##"); break;
					case Hit:   SetAttr(0x47); printf((Kentta[0][x][y]&Idle)?"/\\":"@@"); break;
					case Try:   SetAttr(0x18); printf("/\\"); break;
				}
	}
	fflush(stdout);
}

static void Terminate(int Error)
{
	printf("\33[m\33[23H\r");
	
	if(Error & 0x40)
	{
		printf("Connection terminated. %s ", (Error&0x80)?"You":"Another end");
		switch(Error & ~0xC0)
		{
			case SIGFPE:
				puts("found a Floating Point Exception.");
				break;
			case SIGSEGV:
				puts("found a Segmentation Fault.");
				break;
			case SIGQUIT:
			case SIGINT:
			case SIGABRT:
				puts("terminated quite normally.");
				break;
			default:
				printf("caused error %02X\n", Error);
		}
	}
	else
		printf(
			"Strange termination with error code %d (%s)\n", 
			errno, sys_errlist[errno]);
	
	close(Sock);	
}

static void TermHandler(int SigNum)
{
	char Error[64];
	printf("\33[m\33[23H\r");
	memset(Error, Idle, sizeof Error);
	Error[62] = End;
	Error[63] = SigNum;	
	SafeSend(Sock, Error, sizeof Error);
	Terminate(SigNum|0xC0);
	donecurses();
	exit(EXIT_FAILURE);
}

/*
	Laivaprotokolla:
	  case (byte What) of
	    end  : (byte signal);
	    try  : (byte x, y);
	    hit  : (byte x, y);
	    uppo : (byte x, y, signed byte Len);
	    idle : ();
	    vuoro: ();
*/

static int PlayGame(int MyTurn)
{
	static char Pituus[ShipCount] = {1,1,1,1, 2,2,2, 3,3, 4, 1,1,1};
	int a;
	
	printf("\33[m\33[2J\33[H");
	SetAttr(7);
	
	memset(Oma,    Tyhja, sizeof Oma);
	memset(Toinen, Tyhja, sizeof Toinen);
	
	signal(SIGINT, TermHandler);
	signal(SIGQUIT,TermHandler);
	signal(SIGABRT,TermHandler);
	signal(SIGSEGV,TermHandler);
	signal(SIGFPE, TermHandler);
	
	for(a=0; a<ShipCount; a++)
	{
		int b;
		
Retry:	Laivani[a].x = Random(10);
		Laivani[a].y = Random(10);
		
		if(Random(2))
		{
			Laivani[a].LenDir = -Pituus[a];
			
			if(Laivani[a].y + Pituus[a] > 10)goto Retry;
			
			for(b=0; b<Pituus[a]; b++)
			{
				int tx = Laivani[a].x;
				int ty = Laivani[a].y+b;
				if(Oma[tx][ty] != Tyhja)goto Retry;
				if(tx>0)if(Oma[tx-1][ty] != Tyhja)goto Retry;
				if(tx<9)if(Oma[tx+1][ty] != Tyhja)goto Retry;
				if(ty>0)if(Oma[tx][ty-1] != Tyhja)goto Retry;
				if(ty<9)if(Oma[tx][ty+1] != Tyhja)goto Retry;
			}
			
			for(b=0; b<Pituus[a]; b++)
				Oma[Laivani[a].x][Laivani[a].y+b] = Laiva;
		}
		else
		{
			Laivani[a].LenDir = Pituus[a];
			if(Laivani[a].x + Pituus[a] > 10)goto Retry;
			
			for(b=0; b<Pituus[a]; b++)
			{
				int tx = Laivani[a].x+b;
				int ty = Laivani[a].y;
				if(Oma[tx][ty] != Tyhja)goto Retry;
				if(tx>0)if(Oma[tx-1][ty] != Tyhja)goto Retry;
				if(tx<9)if(Oma[tx+1][ty] != Tyhja)goto Retry;
				if(ty>0)if(Oma[tx][ty-1] != Tyhja)goto Retry;
				if(ty<9)if(Oma[tx][ty+1] != Tyhja)goto Retry;
			}
			
			for(b=0; b<Pituus[a]; b++)
				Oma[Laivani[a].x+b][Laivani[a].y] = Laiva;
		}
	}
	
	
	SetAttr(12); 
	printf("\33[22;5HYour ships are here");
	printf("\33[22;35HOpponent's ships are here");
	
	UpdateField(&Oma,     5,2, -1,-1, 9,9);
	UpdateField(&Toinen, 35,2, -1,-1, 9,9);
	
	SetAttr(15);   printf( "\33[1;58HLegend:");
	
	SetAttr(0x13); printf( "\33[3;58H%%%%");
	               printf( "\33[4;58H%%%%"); SetAttr(14); printf(" Uponnut");
	
	SetAttr(0x78); printf( "\33[6;58H##");
	               printf( "\33[7;58H##"); SetAttr(14); printf(" Turvassa viel");
	
	SetAttr(0x47); printf( "\33[9;58H@@");
	               printf("\33[10;58H@@"); SetAttr(14); printf(" Thn on osuttu");
	               
	SetAttr(0x1B); printf("\33[12;58H ~");
	SetAttr(0x13); printf("\33[13;58H~ "); SetAttr(14); printf(" Avomerta ehk");
	
	SetAttr(0x09); printf("\33[15;58H\\/");
	               printf("\33[16;58H/\\"); SetAttr(14); printf(" Rasti");
	
	for(;;)
	{
		unsigned char What;
		  signed char LenDir;
		unsigned char x,y, CursX=0,CursY=0;
		int Skippaa = 0;
				
		Field Temp;
	
		if(MyTurn)goto Mina;
		
		SetAttr(0x8A);
		printf("\33[23H\rHold on...\33[K\a"); fflush(stdout);
		
		for(;;)
		{
			switch(What = WantByte())
			{
				case Vuoro:
					if(Skippaa)
					{
						Skippaa--;
						break;
					}
					else
						goto Mina;
				case End: 
				{
					x = WantByte();
					return x | 0x40;
				}
				case Try:
					CursX = x = WantByte();
					CursY = y = WantByte();
					
					if((Oma[x][y] == Laiva)||(Oma[x][y] == Hit))
					{
						int a, b;
						for(a=0; a<ShipCount; a++)
						{
							int Osui = 0;
							if(Laivani[a].LenDir<0)
							{
								for(b = 0; b < -Laivani[a].LenDir; b++)
									if((Laivani[a].x   == x)
									 &&(Laivani[a].y+b == y))
									{
										Osui = 1;
										break;
									}
							}
							else
							{
								for(b = 0; b < Laivani[a].LenDir; b++)
									if((Laivani[a].x+b == x)
									 &&(Laivani[a].y   == y))
									{
										Osui = 1;
										break;
									}
							}
							if(Osui)
							{
								Oma[x][y] = What = Hit;
								SafeSend(Sock, &Oma[x][y], 1);
								SafeSend(Sock, &x, 1);
								SafeSend(Sock, &y, 1);								
								
								if(Laivani[a].LenDir<0)
								{
									for(b = 0; b < -Laivani[a].LenDir; b++)
										if(Oma[Laivani[a].x][Laivani[a].y+b]==Laiva)goto Eips;
								}
								else
								{
									for(b = 0; b < Laivani[a].LenDir; b++)
										if(Oma[Laivani[a].x+b][Laivani[a].y]==Laiva)goto Eips;
								}
								if(Laivani[a].LenDir<0)
									for(b = 0; b < -Laivani[a].LenDir; b++)
										Oma[Laivani[a].x][Laivani[a].y+b] = Uppo;
								else
									for(b = 0; b < Laivani[a].LenDir; b++)
										Oma[Laivani[a].x+b][Laivani[a].y] = Uppo;
										
								UpdateField(&Oma, 5,2, -1,-1, 9,9);
								
								SafeSend(Sock, &Oma[x][y], 1);
								SafeSend(Sock, &Laivani[a], 3);
Eips:;						}
						}
						What = Vuoro;
						SafeSend(Sock, &What, 1);
						Skippaa++;
					}
					if(Oma[x][y] == Tyhja)Oma[x][y] = Try;
					UpdateField(&Oma, 5,2, x,y, 9,9);
					break;
				case Hit:
					x = WantByte();
					y = WantByte();
					Toinen[x][y] = Hit;
					UpdateField(&Toinen, 35,2, x,y, x,y);
					break;
				case Uppo:
					x = WantByte();
					y = WantByte();
					LenDir = WantByte();
					
					if(LenDir < 0)
					{
						int b = -LenDir-1;
						for(; b>=0; b--)
						{
							Toinen[x][y+b] = Uppo;
							UpdateField(&Toinen, 35,2, x,y+b, x,y+b);
						}
					}
					else
					{
						int b = LenDir-1;
						for(; b>=0; b--)
						{
							Toinen[x+b][y] = Uppo;
							UpdateField(&Toinen, 35,2, x+b,y, x+b,y);
						}
					}
				case Idle:
					break;
			}
			x=y=LenDir=What;
		}
		
Mina:	MyTurn = 1;

		SetAttr(0x0A);
		printf("\33[23H\rIt's your turn - place your bomb\a");
		fflush(stdout);
			
		while(MyTurn)
		{
			unsigned char ch;
			
			memcpy(&Temp, &Toinen, sizeof Temp);
			
			Temp[CursX][CursY] |= Idle;
			UpdateField(&Temp, 35,2, CursX,CursY, CursX,CursY);
			
			fread(&ch, 1, 1, stdin);
			
			UpdateField(&Toinen, 35,2, CursX,CursY, CursX,CursY);
			
			if((ch=='')||(ch=='A')||(ch=='8'))CursY = (CursY + 9)%10;
			if((ch=='')||(ch=='B')||(ch=='2'))CursY = (CursY + 1)%10;
			if((ch=='')||(ch=='D')||(ch=='4'))CursX = (CursX + 9)%10;
			if((ch=='')||(ch=='C')||(ch=='6'))CursX = (CursX + 1)%10;
			if((ch=='Q')||(ch=='q'))raise(SIGINT);
			if((ch=='\n')||(ch==' '))
				if(Toinen[CursX][CursY] == Tyhja)
					MyTurn=0;
			
		}
		
		Toinen[CursX][CursY] = Try;
		UpdateField(&Toinen, 35,2, CursX,CursY, CursX,CursY);
		SafeSend(Sock, &Toinen[CursX][CursY], 1);
		SafeSend(Sock, &CursX, 1);
		SafeSend(Sock, &CursY, 1);
		CursX = Vuoro;
		SafeSend(Sock, &CursX, 1);
	}
}

static void SubMakeConnection(void)
{
	printf("Waiting for connection in port %d...\n", Port);
}

static int MakeConnection(void)
{
	if(HostName)
	{
		printf("Contacting host %s in port %d...\n", HostName, Port);
		Sock = ConnectOut(HostName, Port);
		return 1; /* Yep, your turn first */
	}
	else
	{
		struct sockaddr_in TheirAddr;
		Sock = WaitForConnection(&Port, &TheirAddr, SubMakeConnection);
		return 0; /* No it is not your turn */
	}
}

int main(int argc, const char *const *argv)
{
	int ShowHelp = 0;
	int Error = 0;
	
	while(--argc)
	{
		switch(*argv[argc])
		{
			case '-':
			case '/':
			{
				const char *s = argv[argc];
				while(s[1])
				{
					switch(s[1])
					{
						case '-':
						case '/':
							break;
						case 'H':
						case 'h':
						case '?': ShowHelp = 1; break;
						default:
							printf("Invalid option: %c\n", s[1]);
							Error = 1;
					}
					s++;
				}
				break;
			}
			default:
				if(strchr(argv[argc], '.'))
				{
GetHostName:		if(HostName != NULL)
					{
						printf("Host name (%s) was already specified.\n", HostName);
						Error = 1;
					}
					else
						HostName = argv[argc];
				}
				else
				{
					char *EP;
					int a;
					a = strtol(argv[argc], &EP, 10);
					if(*EP)
					{
						if(EP != argv[argc])
							printf("Port number ends with garbage (%s). Maybe it is host name.\n", EP);
						goto GetHostName;
					}
					if((Port = a) == 0)
					{
						printf("Port number is strange.\n");
						Error = 1;
					}
				}
		}
	}
	if(!Port)
		printf("Now we don't have a sensitive port number. Let's try default (%d).\n", 
			Port=DefaultPort);
			
	if(Error)
	{
		printf("Errors found. See '%s -h' for help.\n", argv[0]);
		if(!ShowHelp)return Error;
	}
	if(ShowHelp)
	{
		printf(
			"Usage: %s [-Hh?] [host] port\n"
			"\nOptions:                      Host examples:   Port example:"
			"\n  -h  shows this help         194.100.73.1     3000"
			"\n  -H  shows this help too     clinet.fi"
			"\n  -?  also this help page"
			"\n"
			"\nPort value used, if default, is %d\n", argv[0], DefaultPort);
		return 0;
	}
	
	Randomize();
	
	initcurses();
	Terminate(PlayGame(MakeConnection()));
	
	return 0;
}
